package de.intevation.flys.client.server;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import javax.xml.xpath.XPathConstants;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.apache.log4j.Logger;

import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import de.intevation.artifacts.common.ArtifactNamespaceContext;
import de.intevation.artifacts.common.utils.XMLUtils;

import de.intevation.artifacts.httpclient.exceptions.ConnectionException;
import de.intevation.artifacts.httpclient.http.HttpClient;
import de.intevation.artifacts.httpclient.http.HttpClientImpl;

import de.intevation.flys.client.shared.exceptions.ServerException;
import de.intevation.flys.client.client.services.WQInfoService;
import de.intevation.flys.client.shared.model.WQInfoObject;
import de.intevation.flys.client.shared.model.WQInfoObjectImpl;


/**
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class WQInfoServiceImpl
extends      RemoteServiceServlet
implements   WQInfoService
{
    private static final Logger logger =
        Logger.getLogger(WQInfoServiceImpl.class);

    public static final String ERROR_NO_WQINFO_FOUND =
        "error_no_wqinfo_found";

    public static final String XPATH_WQS = "art:service/art:mainvalues/art:mainvalue";


    public WQInfoObject[] getWQInfo(
        String locale,
        String river,
        double from,
        double to)
    throws ServerException
    {
        logger.info("WQInfoServiceImpl.getWQInfo");

        String url = getServletContext().getInitParameter("server-url");

        Document doc = XMLUtils.newDocument();

        XMLUtils.ElementCreator ec = new XMLUtils.ElementCreator(
            doc,
            ArtifactNamespaceContext.NAMESPACE_URI,
            ArtifactNamespaceContext.NAMESPACE_PREFIX);

        Element mainvalues = ec.create("mainvalues");
        Element riverEl = ec.create("river");
        Element startEl = ec.create("start");
        Element endEl   = ec.create("end");

        riverEl.setTextContent(river);
        startEl.setTextContent(Double.valueOf(from).toString());
        endEl.setTextContent(Double.valueOf(to).toString());

        mainvalues.appendChild(riverEl);
        mainvalues.appendChild(startEl);
        mainvalues.appendChild(endEl);

        doc.appendChild(mainvalues);

        HttpClient client = new HttpClientImpl(url, locale);

        try {
            Document result = client.callService(url, "mainvalues", doc);

            logger.debug("Extract wq info objects now.");
            WQInfoObject[] objects = extractWQInfoObjects(result);

            if (objects.length > 0) {
                return objects;
            }
        }
        catch (ConnectionException ce) {
            logger.error(ce, ce);
        }

        throw new ServerException(ERROR_NO_WQINFO_FOUND);
    }


    /**
     * Extracts all wq info objects from <i>result</i> document.
     *
     * @param result The document retrieved by the server.
     *
     * @return a list of WQInfoObjects.
     */
    protected WQInfoObject[] extractWQInfoObjects(Document result)
    throws    ServerException
    {
        NodeList list = (NodeList) XMLUtils.xpath(
            result,
            XPATH_WQS,
            XPathConstants.NODESET,
            ArtifactNamespaceContext.INSTANCE);

        if (list == null || list.getLength() == 0) {
            logger.warn("No wq info found.");

            throw new ServerException(ERROR_NO_WQINFO_FOUND);
        }

        boolean debug = logger.isDebugEnabled();

        int num = list.getLength();
        if (debug) {
            logger.debug("Response contains " + num + " objects.");
        }

        List<WQInfoObject> objects =
            new ArrayList<WQInfoObject>(num);

        for (int i = 0; i < num; i++) {
            WQInfoObject obj = buildWQInfoObject((Element)list.item(i));

            if (obj != null) {
                objects.add(obj);
            }
        }

        if (debug) {
            logger.debug("Retrieved " + objects.size() + " wq values");
        }

        WQInfoObject [] array = (WQInfoObject[])
            objects.toArray(new WQInfoObject[objects.size()]);

        Arrays.sort(array, WQ_INFO_OBJECT_CMP);

        return array;
    }

    public static final Comparator<WQInfoObject> WQ_INFO_OBJECT_CMP =
        new Comparator<WQInfoObject>() {
            @Override
            public int compare(WQInfoObject a, WQInfoObject b) {

                // Descending by type: Qs before Ds
                int cmp = a.getType().compareTo(b.getType());
                if (cmp < 0) return +1;
                if (cmp > 0) return -1;

                // Ascending by value
                double diff = a.getValue() - b.getValue();
                if (diff < 0d) return -1;
                if (diff > 0d) return +1;
                return 0;
            }
        };

    /**
     * Extracts information for a single wq info object and intializes an
     * WQInfoObject with them.
     *
     * @param node The node that contains the information.
     *
     * @return a valid WQInfoObject.
     */
    protected static WQInfoObject buildWQInfoObject(Element node) {

        String uri = ArtifactNamespaceContext.NAMESPACE_URI;

        String name  = node.getAttributeNS(uri, "name");
        String type  = node.getAttributeNS(uri, "type");
        String value = node.getAttributeNS(uri, "value");

        if (!name.isEmpty() && !type.isEmpty()) {
            try {
                return new WQInfoObjectImpl(
                    name,
                    type,
                    new Double(value));
            }
            catch (NumberFormatException nfe) {
                logger.warn(nfe.getLocalizedMessage());
            }
        }

        logger.warn("Invalid wq info object found.");

        return null;
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
